/*
** Author:      James Painter
** Purpose:     2 parameter family of cubic filters
** Reference:   "Reconstruction Filters in Computer Graphics",
**              Don Mitchell, Arun Netravali  AT&T Bell Laboratories
**              SIGGraph `88.
**
** Copyright (c), 1988 GRAIL, University of Washington
**
** $Revision: 1.1 $
** $Date: 89/01/25 20:17:07 $
** $Locker: jamie $
** $Log:	CubicFilter.c,v $
 * Revision 1.1  89/01/25  20:17:07  jamie
 * Initial revision
 * 
 * Revision 1.1  89/01/11  21:27:08  jamie
 * Initial revision
 * 
*/

#include <stdio.h>
#include <math.h>

#ifndef _POLY4
#  define _POLY1(x, c)	((c)[0] * (x) + (c)[1])
#  define _POLY2(x, c)	(_POLY1((x), (c)) * (x) + (c)[2])
#  define _POLY3(x, c)	(_POLY2((x), (c)) * (x) + (c)[3])
#  define _POLY4(x, c)	(_POLY3((x), (c)) * (x) + (c)[4])
#endif

/* Indefinite integral of the filter polynomials
 * Polynomials are stored int the power basis X^4,X^3,X^2,X,1 (in that order)
 */
static double poly1[5], poly2[5]; 

/*
**  Initilize the cubic filter.  
**  Sets up the indefinite integrals of the filter polynomials.
*/
CubicFilterSetup( B, C)
     double B, C;		/* Controlling parameters to select which
				 * cubic from the family.
				 */
{
  int i;
  double factor;

  /*  The filter is defined by:
  **  
  **    ( (12-9B-6C)|x|^3   + (-18 + 12B + 6C)|x|^2 + (6-2B) )  / 6.0
  **             when |x| < 1
  **  
  **    ( (-B-6C)|x|^3 +(6B+30C)|x|^2 + (-12b - 48C)|x| + (8B + 24C) ) / 6.0
  **             when 1 <= |x| < 2
  **  
  **    0 otherwise
  */

  /* 
  **  We form the indefinite integral polynomials for later evaluation
  **  of definite integrals over intervals.
  */
  poly1[0] = ( 12.0 -  9.0*B -  6.0*C) / 4.0;   /* X^4 coefficient */
  poly1[1] = (-18.0 + 12.0*B +  6.0*C) / 3.0;   /* X^3 coefficient */
  poly1[2] = 0;                                 /* X^2 coefficient */
  poly1[3] = (  6.0 - 2.0*B         );          /* X^1 coefficient */
  poly1[4] = 0.0;                               /* X^0 coefficient */

  poly2[0] = (            -B -  6.0*C) / 4.0;   /* X^4 coefficient */
  poly2[1] = (         6.0*B + 30.0*C) / 3.0;   /* X^3 coefficient */
  poly2[2] = (       -12.0*B - 48.0*C) / 2.0;   /* X^2 coefficient */
  poly2[3] = (         8.0*B + 24.0*C)      ;   /* X^1 coefficient */
  poly2[4] = 0.0;                               /* X^0 coefficient */

  /* Actually, these are all scaled by 1.0/6.0 */
  factor = 1.0 / 6.0;
  for(i=0; i< 5; i++) {
    poly1[i] *=  factor;  poly2[i] *= factor;
  }
}

/* Evaluate the (1-D) integral over the interval x1 to x2. */
static double
Integrate1D( x1, x2 )
     double x1, x2;
{
  double result = 0.0;
  double sign, tmp;

  /* Bring the interval into the range where the function is non-zero.
  **  The integral of any area outside this range must be zero.
  */
  if (x1 < -2.0) x1 = -2.0;
  if (x2 < -2.0) x2 = -2.0;
  if (x1 > 2.0)  x1 = 2.0;
  if (x2 > 2.0)  x2 = 2.0;

  /* Be sure that x1 <= x2 */
  if (x1 > x2) {
    sign = -1.0;
    tmp = x1;
    x1  = x2;
    x2  = tmp;
  } else {
    sign = 1.0;
  }

  /* Find the contribution in the range -2.0 to -1.0 */
  if (x1 <= -1.0) {
    if (x2 <= -1.0) {
      result += ( - _POLY4( -x2,poly2) + _POLY4(-x1,poly2) );
      return sign*result;
    } else
      result += ( - _POLY4( 1.0, poly2) + _POLY4(-x1, poly2) );
    x1 = -1.0;
  }
  if (x1 < 0.0) {
    if (x2 <= 0.0) {
      result += ( - _POLY4( -x2,poly1) + _POLY4(-x1,poly1) );
      return sign*result;
    } else
      result += ( - _POLY4(0.0, poly1) + _POLY4(-x1, poly1) );
    x1 = 0.0;
  }
  if (x1 < 1.0) {
    if (x2 < 1.0) {
      result += (   _POLY4( x2,poly1) - _POLY4(x1,poly1) );
      return sign*result;
    } else
      result += (   _POLY4(1.0,poly1) - _POLY4(x1, poly1) );
    x1 = 1.0;
  }
  result += ( _POLY4( x2, poly2) - _POLY4(x1,poly2) );
  return sign*result;
}


/* 
**  Lookup the result of the 2-D integral over the rectangle define by
**  (xmin,ymin)  to (xmax,ymax).
*/
double
IntegralLookup( xmin, ymin, xmax, ymax )
     double xmin, ymin, xmax, ymax;
{
  /* Use the separability of the filter to reduce to two 1-D problems. */
  return Integrate1D(xmin,xmax) * Integrate1D(ymin,ymax);
}


  



